<?php
/* vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4: */
/**
 * XML_Indexing reader class
 *
 * PHP versions 4 and 5
 *
 * LICENSE: This source file is subject to version 3.0 of the PHP license
 * that is available through the world-wide-web at the following URI:
 * http://www.php.net/license/3_0.txt.  If you did not receive a copy of
 * the PHP License and are unable to obtain it through the web, please
 * send a note to license@php.net so we can mail you a copy immediately.
 *
 * @category   XML
 * @package    XML_Indexing
 * @copyright  2004 Samalyse SARL corporation
 * @author     Olivier Guilyardi <olivier@samalyse.com>
 * @license    http://www.php.net/license/3_0.txt  PHP License
 * @version    CVS: $Id:$
 * @link       http://pear.php.net
 * @since      File available since Release 0.1
 */
    
/**
 * Transparent XML Indexing Reader
 *
 * This class allows to work on big XML files without madly increasing access time.
 * For this purpose, it creates an index, which contains informations to rapidly
 * seek through a given XML file to retrieve a specific portion of it.
 * 
 * The indexing process is based on XPath expressions. Not all of the XPath language,
 * but an appropriate subset for what a big XML files is expected to contain.
 *
 * Currently, this class works transparently, creating specific indexes upon specific
 * requests.
 *
 * For example, when initially looking for /foo/bar[232], all instances from
 * /foo/bar[1] to /foo/bar[n] will get indexed (so the first run is slow). 
 * Subsequent calls with such expressions as /foo/bar[232], /foo/bar[100], 
 * /foo/bar[25], etc... will then all make use of the created index (fast).
 * 
 * In addition to numerical indexes, attribute values indexing is currently
 * supported as well. That is, expressions as /foo/bar[@id='someValue']. Similarly
 * to the numerical indexing process, looking for a such expression will index
 * all values of the 'id' attribute for the given XPath root (/foo/bar here).
 *
 * Using this class is pretty straightforward :
 *
 * <code>
 * $reader = new XML_Indexing_Reader ('test.xml');
 * $reader->find('/foo/bar[232]'); // Or any other XPath expression
 * 
 * echo "Extracted XML data : "
 * while ($xml = $reader->next()) {
 *     echo $xml;
 * }
 * </code>
 *
 * Namespaces extraction is supported. These namespaces declarations are stored
 * in the index files. You can retrieve them with :
 *
 * <code>
 * $reader->find(...); // Needs to be call prior to getNamespaces()
 * $nsList = $reader->getNamespaces();
 * foreach ($nsList as $prefix => $uri) {
 *     echo "$prefix => $uri";
 * }
 * </code>
 * 
 * The index storage strategy can be customized by modifying the default dsn
 * value. Currently, only local file containers are supported.
 * <code>
 *
 * // The following will store indexes in /tmp, using file names with an .xi 
 * // prefix. That is the default.
 * $options['dsn'] = 'file:///tmp/%s.xi';
 * $indexer = new XML_Indexing_Reader ('test.xml', $options);
 *
 * // You can specify your own path as long as you include the %s expression :
 * $options['dsn'] = 'file:///var/cache/xi/%s.xi'
 * $indexer = new XML_Indexing_Reader ('test.xml', $options);
 * </code>
 *
 * See the constructor documentation for more information on options.
 *
 * @copyright  2004 Samalyse SARL corporation
 * @author     Olivier Guilyardi <olivier@samalyse.com>
 * @license    http://www.php.net/license/3_0.txt  PHP License
 * @version    Release: @package_version@
 * @link       http://pear.php.net
 * @since      Class available since Release 0.1
 */
class XML_Indexing_Reader {
   
    /**
     * Global options
     * @var array
     * @access private
     */
    var $_options = array (
            'dsn' => null, // dynamically generated
            'gz_level' => 0,
        );
  
    /**
     * Currently used index
     * @var array
     * @access private
     */
    var $_index = array();

    /**
     * XML file name being parsed
     * @var string
     * @access private
     */
    var $_xmlFilename = null;
  
    /**
     * XML file resource
     * @var resource
     * @access private
     */
    var $_xmlFileRes = null;
   
    /**
     * Matching regions of the XML file
     * 
     * @var array
     * @access private
     */
     var $_regions = array();

     /**
      * Counter as used by next()
      *
      * @var int
      * @access private
      * @see XML_Indexing_Reader::next()
      */
      var $_fetchCounter;
    
    /**
     * Constructor
     * 
     * Supported options : 
     *
     * - "dsn" : Index storage strategy, Default is to create a file in the
     *           system default temporary directory (ie: /tmp on *nix),
     *           with a '.xi' suffix.
     *           The only currently supported format is 'file://<path>'.
     *           Example : 'file:///var/cache/xi/%s.xi'
     *           Using the '%s' expression is required.
     * - "gz_level" : Zlib compression level of the index files. 0 by default
     *                (no compression). Goes up to 9 (maximum compression, slow).
     *                Use this if you expect big indexes (many attributes, etc...)
     *
     * @param string $filename The XML file to parse
     * @param array  $options  Optional custom options 
     * @access private 
     */
    function XML_Indexing_Reader ($filename, $options = array()) 
    {
        $this->_options = array_merge ($this->_options, $options);
        if (is_null($this->_options['dsn'])) {
            $tmpdir = dirname(tempnam('/tmp','junk'));
            $this->_options['dsn'] = "file://$tmpdir/%s.xi";
        }
        $this->_xmlFilename = $filename;
        $this->_loadIndex ();
    }
    
    /**
     * Load the index attached to the current XML file
     * 
     * @access private
     * @return void
     * @see XML_Indexing_Reader::_xmlFilename
     * @see XML_Indexing_Reader::_index
     */
    function _loadIndex () 
    {

        $index = array();
        
        $filename = realpath ($this->_xmlFilename);
        
        $stat = stat ($filename);
        
        $indexName = md5($filename) . '.' . 
                     md5($stat['ctime'] . ':' . $stat['size']) .
                     ($this->_options['gz_level'] ? '.gz' : '');
        
        $a = strcspn ($this->_options['dsn'], ':');
        $proto = substr ($this->_options['dsn'], 0, $a);
        $path = substr ($this->_options['dsn'], $a + 3);

        switch ($proto) {
            case 'file' : 
                $f = sprintf($path, $indexName);
                echo "Loading $f\n";
                $data = '';
                if ($this->_options['gz_level']) {
                    if ($zp = @gzopen ($f, 'rb')) {
                        while (!gzeof($zp)) {
                            $data .= gzread ($zp, 0xFFFF);
                        }
                        gzclose($zp);
                    }
                } else {
                    $data = @file_get_contents ($f);
                }

                if ($data) {
                    $index = unserialize($data);
                }
                break;
        }
        $this->_index = $index;
    }
    
    /**
     * Save the current index
     *
     * @access private
     * @return void
     * @see XML_Indexing_Reader::_xmlFilename
     * @see XML_Indexing_Reader::_index
     */
    function _saveIndex ()
    {
        $filename = realpath ($this->_xmlFilename);
        
        $stat = stat ($filename);
        
        $indexName = md5($filename) . '.' . 
                     md5($stat['ctime'] . ':' . $stat['size']) .
                     ($this->_options['gz_level'] ? '.gz' : '');
        
        $a = strcspn ($this->_options['dsn'], ':');
        $proto = substr ($this->_options['dsn'], 0, $a);
        $path = substr ($this->_options['dsn'], $a + 3);

        switch ($proto) {
            case 'file' : 
                $f = sprintf($path, $indexName);
                $data = serialize($this->_index);
                if ($this->_options['gz_level']) {
                    if ($zp = gzopen ($f, 'wb' . $this->_options['gz_level'])) {
                        gzwrite ($zp, $data);
                        gzclose ($zp);
                    }
                } else {
                    $fp = fopen ($f,'w');
                    fwrite ($fp, $data);
                    fclose ($fp);
                }
                break;
        }
    }

    /**
     * Build and save an Index
     * 
     * @param string $type The type of index
     * @param string $root XPath root
     * @param string $attr Optional attribute name
     * @access private
     * @return void
     */
    function _buildIndex ($type, $root, $attr = null)
    {
        include_once "XML/Indexing/Builder/$type.php";
        switch ($type) {
            case 'Numeric':
                $builder = new XML_Indexing_Builder_Numeric ($this->_xmlFilename, 
                                                             $root);
                $this->_index[$root]['#'] = $builder->getIndex();
                break;
            case 'Attribute':
                $builder = new XML_Indexing_Builder_Attribute ($this->_xmlFilename, 
                                                               $root, $attr);
                $this->_index[$root][$attr] = $builder->getIndex();
                break;
        }
        $this->_index['NS'] = $builder->getNamespaces();
        $this->_saveIndex();
    }
    
    /**
     * Search for an XPath expression
     * 
     * @param string $xpath XPath expression to look for
     * @return bool True if found, false otherwise, or a PEAR_Error in case of 
     *              failure
     * @access public
     */
    function find ($xpath) 
    {
        $this->_regions = array();
        $this->_fetchCounter = 0;
        if (ereg('^([a-zA-Z0-9:._/-]+)\[(.*)\]$',$xpath, $regs)) {
            $root = $regs[1];
            $expr = $regs[2];
            if (!$this->_xmlFileRes) {
                if (!$this->_xmlFileRes = fopen ($this->_xmlFilename, 'r'))
                return new PEAR_Error('Unable to open the XML File : ' . 
                                      $this->_xmlFilename);
            }
            if (is_numeric($expr)) {
                if (!isset($this->_index[$root]['#'])) {
                    $this->_buildIndex ('Numeric',$root);
                }
                if (isset($this->_index[$root]['#'][$expr])) {
                    foreach ($this->_index[$root]['#'][$expr] as $spec) {
                        $this->_regions[] = $spec;
                    }
                }
            } else if (ereg('^@([a-zA-Z0-9:._-]+)=[\'"](.*)[\'"]$', $expr ,$regs)) {
                $attr = $regs[1];
                $value = stripcslashes($regs[2]);
                if (!isset($this->_index[$root][$attr])) {
                    $this->_buildIndex ('Attribute', $root, $attr);
                }
                if (isset($this->_index[$root][$attr][$value])) {
                    foreach ($this->_index[$root][$attr][$value] as $spec) {
                        $this->_regions[] = $spec;
                    }
                }
            } else if (ereg('^@([a-zA-Z0-9:._-]+)$', $expr ,$regs)) {
                $attr = $regs[1];
                if (!isset($this->_index[$root][$attr])) {
                    $this->_buildIndex ('Attribute', $root, $attr);
                }
                if (!empty($this->_index[$root][$attr])) {
                    foreach ($this->_index[$root][$attr] as $value => $regions) {
                        foreach ($regions as $spec) {
                            $this->_regions[] = $spec;
                        }
                    }
                }
            }
        }
        return !empty($this->_regions);
    }

    /**
     * Retrieves the number of matches
     *
     * @return int The number of matches
     * @access public
     */
    function count()
    {
        return count($this->_regions);
    }

    /**
     * Fetch the next raw XML data portion
     * 
     * @return string Current portion of XML data or an empty string if 
     *                none left
     * @access public
     */
     function next() 
     {
       $xmlStr = '';  
       if (isset($this->_regions[$this->_fetchCounter])) {
           list ($ofs, $len) = $this->_regions[$this->_fetchCounter++];
           fseek ($this->_xmlFileRes, $ofs);
           $xmlStr = fread ($this->_xmlFileRes, $len);
       }
       return $xmlStr;
         
     }

     /**
      * Return namespaces declared in the XML file
      * 
      * @return array An associative array of the form ('prefix' => 'uri', ...)
      * @access public
      */
     function getNamespaces ()
     {
         if (isset($this->_index['NS'])) return $this->_index['NS'];
         else return array();
     }
}

?>
